#ifndef _DBUTILS_DAQCONFIG_GENERALDATAREADER_H
#define _DBUTILS_DAQCONFIG_GENERALDATAREADER_H

#include <NovaDatabase/Table.h>
#include <boost/lexical_cast.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/shared_ptr.hpp>
#include <string>
#include <vector>
#include <map>

namespace NOVADB = nova::database;

namespace dbutils {
namespace daqconfig {

template <class T, class U>
struct SimpleDataContainer {
 public:
    T key;
    U value;
    bool hasValue;
};

template <class T>
struct SimpleDataSet
{
    uint32_t numberOfValidValues;
    std::map<uint32_t, boost::shared_ptr<SimpleDataContainer<uint32_t, T> > > dataMap;
    bool isComplete() {return (numberOfValidValues == dataMap.size());}
};

/**
 * Class to handle fetching of DAQ configuration data from the database.
 *
 * @author Kurt Biery
 * @version $Revision: 1.10 $ $Date: 2011/11/01 20:22:21 $
 */

class GeneralDataReader {

 public:

    // ********************
    // *** Constructors ***
    // ********************

    GeneralDataReader(const std::string& tableDefinitionFilename,
                      const uint64_t& subsystemConfigId,
                      const std::string* namesTableDefinitionFilename = 0);

    GeneralDataReader(const std::string& tableDefinitionFilename,
                      const uint64_t& subsystemConfigId,
                      const std::string& primarySelectionColumnName,
                      const std::string* namesTableDefinitionFilename = 0);

    GeneralDataReader(const std::string& tableDefinitionFilename,
                      const uint64_t& subsystemConfigId,
                      const std::string& primarySelectionColumnName,
                      const std::string& secondarySelectionColumnName,
                      const std::string* namesTableDefinitionFilename = 0);

    GeneralDataReader(const std::string& tableDefinitionFilename,
                      const uint64_t& subsystemConfigId,
                      const std::string& primarySelectionColumnName,
                      const std::string& secondarySelectionColumnName,
                      const std::string& tertiarySelectionColumnName,
                      const std::string* namesTableDefinitionFilename = 0);


    GeneralDataReader(const std::string& tableDefinitionFilename,
                      const std::vector<int64_t>& subsystemConfigIdList,
                      const std::string* namesTableDefinitionFilename = 0);

    GeneralDataReader(const std::string& tableDefinitionFilename,
                      const std::vector<int64_t>& subsystemConfigIdList,
                      const std::string& primarySelectionColumnName,
                      const std::string* namesTableDefinitionFilename = 0);

    GeneralDataReader(const std::string& tableDefinitionFilename,
                      const std::vector<int64_t>& subsystemConfigIdList,
                      const std::string& primarySelectionColumnName,
                      const std::string& secondarySelectionColumnName,
                      const std::string* namesTableDefinitionFilename = 0);

    GeneralDataReader(const std::string& tableDefinitionFilename,
                      const std::vector<int64_t>& subsystemConfigIdList,
                      const std::string& primarySelectionColumnName,
                      const std::string& secondarySelectionColumnName,
                      const std::string& tertiarySelectionColumnName,
                      const std::string* namesTableDefinitionFilename = 0);


    // ****************************
    // *** Fetch a single value ***
    // ****************************

    template<class T>
    bool getValue(const std::string& columnName, T& resultValue);

    template<class T, class U>
    bool getValueFor1Key(const std::string& columnName, T& resultValue,
                         const U& primarySelectionValue);

    template<class T, class U, class V>
    bool getValueFor2Keys(const std::string& columnName, T& resultValue,
                          const U& primarySelectionValue,
                          const V& secondarySelectionValue);

    template<class T, class U, class V, class W>
    bool getValueFor3Keys(const std::string& columnName, T& resultValue,
                          const U& primarySelectionValue,
                          const V& secondarySelectionValue,
                          const W& tertiarySelectionValue);

    // *****************************
    // *** Fetch multiple values ***
    // *****************************

    template<class T, class U>
    bool getListOfValuesFor2Keys(const std::string& columnName,
                                 std::vector<T>& resultValues,
                                 const U& primarySelectionValue,
                                 const uint32_t& secondarySelectionLowValue,
                                 const uint32_t& secondarySelectionHighValue);

    template<class T, class U, class V>
    bool getListOfValuesFor3Keys(const std::string& columnName,
                                 std::vector<T>& resultValues,
                                 const U& primarySelectionValue,
                                 const V& secondarySelectionValue,
                                 const uint32_t& tertiarySelectionLowValue,
                                 const uint32_t& tertiarySelectionHighValue);

    template<class T, class U>
    bool getDataSetFor2Keys(const std::string& columnName,
                            SimpleDataSet<T>& resultDataSet,
                            const U& primarySelectionValue,
                            const uint32_t& secondarySelectionLowValue,
                            const uint32_t& secondarySelectionHighValue);

    template<class T, class U, class V>
    bool getDataSetFor3Keys(const std::string& columnName,
                            SimpleDataSet<T>& resultDataSet,
                            const U& primarySelectionValue,
                            const V& secondarySelectionValue,
                            const uint32_t& tertiarySelectionLowValue,
                            const uint32_t& tertiarySelectionHighValue);

    // ***********************************
    // *** Fetch a hierarchy of values ***
    // ***********************************

    template<class T, class U>
    bool getSetOfValuesFor3Keys(const std::string& columnName,
                                std::vector<boost::shared_ptr<std::vector<T> > >&
                                resultValues,
                                const U& primarySelectionValue,
                                const uint32_t& secondarySelectionLowValue,
                                const uint32_t& secondarySelectionHighValue,
                                const uint32_t& tertiarySelectionLowValue,
                                const uint32_t& tertiarySelectionHighValue);

    template<class T, class U>
    bool getDataSetMapFor3Keys(const std::string& columnName,
                               std::map<uint32_t, boost::shared_ptr<
                               SimpleDataSet<T> > >& twoLevelDataSet,
                               const U& primarySelectionValue,
                               const uint32_t& secondarySelectionLowValue,
                               const uint32_t& secondarySelectionHighValue,
                               const uint32_t& tertiarySelectionLowValue,
                               const uint32_t& tertiarySelectionHighValue);

    // *********************
    // *** General tools ***
    // *********************

    template<class TT>
    static std::vector<std::pair<uint64_t, TT> >
        remapValues(const std::vector<TT>& values,
                    const uint32_t& indexOffset);

 private:

    void _init(const std::string& tableDefinitionFilename,
               const uint64_t& subsystemConfigId,
               const std::string* namesTableDefinitionFilename);

    template<class T>
    bool _getValueForAllConfigs(const std::string& columnName,
                                const std::vector<std::string>& selectionValues,
                                T& resultValue);

    template<class T>
    bool _getValueForSingleConfig(const std::string& columnName,
                                  const uint64_t& configId,
                                  std::vector<std::string> selectionValues,
                                  T& resultValue);

    template<class T>
    bool _getDefaultValue(const std::string& columnName,
                          const uint64_t& configId,
                          const std::vector<std::string>& selectionValues,
                          T& resultValue);

    template<class T>
    bool _getValueFromDB(const std::string& columnName,
                         const uint64_t& configId,
                         const std::vector<std::string>&
                         selectionValues,
                         T& resultValue);

    template<class T>
    bool _getListOfValuesForAllConfigs(const std::string& columnName,
                                       const std::vector<std::string>& selectionValues,
                                       const uint32_t& fineGrainedSelectionLowValue,
                                       const uint32_t& fineGrainedSelectionHighValue,
                                       std::vector<T>& resultValues);

    template<class T> bool
    _getListOfValuesForSingleConfig(const std::string& columnName,
                                   const uint64_t& configId,
                                   std::vector<std::string> selectionValues,
                                   const uint32_t& fineGrainedSelectionLowValue,
                                   const uint32_t& fineGrainedSelectionHighValue,
                                   SimpleDataSet<T>& dataSet);

    template<class T> bool
    _getListOfValuesFromDB(const std::string& columnName,
                           const uint64_t& configId,
                           std::vector<std::string> selectionValues,
                           const uint32_t& fineGrainedSelectionLowValue,
                           const uint32_t& fineGrainedSelectionHighValue,
                           SimpleDataSet<T>& dataSet);

    template<class T>
    bool _getDataSetForAllConfigs(const std::string& columnName,
                                  const std::vector<std::string>& selectionValues,
                                  const uint32_t& fineGrainedSelectionLowValue,
                                  const uint32_t& fineGrainedSelectionHighValue,
                                  SimpleDataSet<T>& resultDataSet);

    template<class T>
    bool _getSetOfValuesForAllConfigs(const std::string& columnName,
                                      const std::vector<std::string>& selectionValues,
                                      const uint32_t& firstLevelSelectionLowValue,
                                      const uint32_t& firstLevelSelectionHighValue,
                                      const uint32_t& secondLevelSelectionLowValue,
                                      const uint32_t& secondLevelSelectionHighValue,
                                      std::vector<boost::shared_ptr<std::vector<T> > >&
                                      resultValues);

    template<class T> bool
    _getSetOfValuesForSingleConfig(const std::string& columnName,
                                  const uint64_t& configId,
                                  std::vector<std::string> selectionValues,
                                  const uint32_t& firstLevelSelectionLowValue,
                                  const uint32_t& firstLevelSelectionHighValue,
                                  const uint32_t& secondLevelSelectionLowValue,
                                  const uint32_t& secondLevelSelectionHighValue,
                                  std::map<uint32_t, boost::shared_ptr<
                                  SimpleDataSet<T> > >& twoLevelDataSet);

    template<class T> bool
    _getSetOfValuesFromDB(const std::string& columnName,
                          const uint64_t& configId,
                          std::vector<std::string> selectionValues,
                          const uint32_t& firstLevelSelectionLowValue,
                          const uint32_t& firstLevelSelectionHighValue,
                          const uint32_t& secondLevelSelectionLowValue,
                          const uint32_t& secondLevelSelectionHighValue,
                          std::map<uint32_t, boost::shared_ptr<
                          SimpleDataSet<T> > >& twoLevelDataSet);

    template<class T>
    bool _getDataSetMapForAllConfigs(const std::string& columnName,
                                     const std::vector<std::string>& selectionValues,
                                     const uint32_t& firstLevelSelectionLowValue,
                                     const uint32_t& firstLevelSelectionHighValue,
                                     const uint32_t& secondLevelSelectionLowValue,
                                     const uint32_t& secondLevelSelectionHighValue,
                                     std::map<uint32_t, boost::shared_ptr<
                                     SimpleDataSet<T> > >& twoLevelDataSet);

    bool _setValidityRanges(const std::vector<std::string>& selectionValues,
                            bool fillInDefaults = true);

    boost::shared_ptr<NOVADB::Table> _dbTablePtr;
    bool _hasValidConnection;

    std::vector<std::string> _selectionColumns;
    std::vector<int64_t> _namedConfigLineage;

    mutable boost::mutex _dataMutex;

    std::string _globalDefault;
};

/**
 * Fetches the requested value...
 */
template<class T>
bool GeneralDataReader::
getValue(const std::string& columnName, T& resultValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    return _getValueForAllConfigs(columnName, selectionValues, resultValue);
}

/**
 * Fetches the requested value...
 */
template<class T, class U>
bool GeneralDataReader::
getValueFor1Key(const std::string& columnName, T& resultValue,
                const U& primarySelectionValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (primarySelectionValue));
    return _getValueForAllConfigs(columnName, selectionValues, resultValue);
}

/**
 * Fetches the requested value...
 */
template<class T, class U, class V>
bool GeneralDataReader::
getValueFor2Keys(const std::string& columnName, T& resultValue,
                 const U& primarySelectionValue,
                 const V& secondarySelectionValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (primarySelectionValue));
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (secondarySelectionValue));
    return _getValueForAllConfigs(columnName, selectionValues, resultValue);
}

/**
 * Fetches the requested value...
 */
template<class T, class U, class V, class W>
bool GeneralDataReader::
getValueFor3Keys(const std::string& columnName, T& resultValue,
                 const U& primarySelectionValue,
                 const V& secondarySelectionValue,
                 const W& tertiarySelectionValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (primarySelectionValue));
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (secondarySelectionValue));
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (tertiarySelectionValue));
    return _getValueForAllConfigs(columnName, selectionValues, resultValue);
}

/**
 * Fetches the requested values...
 */
template<class T, class U>
bool GeneralDataReader::
getListOfValuesFor2Keys(const std::string& columnName,
                        std::vector<T>& resultValues,
                        const U& primarySelectionValue,
                        const uint32_t& secondarySelectionLowValue,
                        const uint32_t& secondarySelectionHighValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (primarySelectionValue));
    return _getListOfValuesForAllConfigs(columnName, selectionValues,
                                         secondarySelectionLowValue,
                                         secondarySelectionHighValue,
                                         resultValues);
}

/**
 * Fetches the requested values...
 */
template<class T, class U, class V>
bool GeneralDataReader::
getListOfValuesFor3Keys(const std::string& columnName,
                        std::vector<T>& resultValues,
                        const U& primarySelectionValue,
                        const V& secondarySelectionValue,
                        const uint32_t& tertiarySelectionLowValue,
                        const uint32_t& tertiarySelectionHighValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (primarySelectionValue));
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (secondarySelectionValue));
    return _getListOfValuesForAllConfigs(columnName, selectionValues,
                                         tertiarySelectionLowValue,
                                         tertiarySelectionHighValue,
                                         resultValues);
}

/**
 * Fetches the requested values...
 */
template<class T, class U>
bool GeneralDataReader::
getDataSetFor2Keys(const std::string& columnName,
                   SimpleDataSet<T>& resultDataSet,
                   const U& primarySelectionValue,
                   const uint32_t& secondarySelectionLowValue,
                   const uint32_t& secondarySelectionHighValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (primarySelectionValue));
    return _getDataSetForAllConfigs(columnName, selectionValues,
                                    secondarySelectionLowValue,
                                    secondarySelectionHighValue,
                                    resultDataSet);
}

/**
 * Fetches the requested values...
 */
template<class T, class U, class V>
bool GeneralDataReader::
getDataSetFor3Keys(const std::string& columnName,
                   SimpleDataSet<T>& resultDataSet,
                   const U& primarySelectionValue,
                   const V& secondarySelectionValue,
                   const uint32_t& tertiarySelectionLowValue,
                   const uint32_t& tertiarySelectionHighValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (primarySelectionValue));
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (secondarySelectionValue));
    return _getDataSetForAllConfigs(columnName, selectionValues,
                                    tertiarySelectionLowValue,
                                    tertiarySelectionHighValue,
                                    resultDataSet);
}

/**
 * Fetches the requested values...
 */
template<class T, class U>
bool GeneralDataReader::
getSetOfValuesFor3Keys(const std::string& columnName,
                       std::vector<boost::shared_ptr<std::vector<T> > >&
                       resultValues,
                       const U& primarySelectionValue,
                       const uint32_t& secondarySelectionLowValue,
                       const uint32_t& secondarySelectionHighValue,
                       const uint32_t& tertiarySelectionLowValue,
                       const uint32_t& tertiarySelectionHighValue)
{
    boost::mutex::scoped_lock sl(_dataMutex);
    std::vector<std::string> selectionValues;
    selectionValues.push_back(boost::lexical_cast<std::string>
                              (primarySelectionValue));
    return _getSetOfValuesForAllConfigs(columnName, selectionValues,
                                        secondarySelectionLowValue,
                                        secondarySelectionHighValue,
                                        tertiarySelectionLowValue,
                                        tertiarySelectionHighValue,
                                        resultValues);
}


template<class TT>
std::vector<std::pair<uint64_t, TT> > GeneralDataReader::
remapValues(const std::vector<TT>& values,
            const uint32_t& indexOffset)
{
    std::vector<std::pair<uint64_t, TT> > resultList;
    for (uint32_t idx = 0; idx < values.size(); ++idx) {
        uint32_t bitIndex = idx + indexOffset;
        TT value = values[idx];
        bool found = false;

        for (uint32_t jdx = 0; jdx < resultList.size(); ++jdx) {
            if (resultList[jdx].second == value) {
                resultList[jdx].first |= ((uint64_t) 1) << bitIndex;
                found = true;
                break;
            }
        }

        if (! found) {
            uint64_t workingBitmap = ((uint64_t) 1) << bitIndex;
            resultList.push_back(std::make_pair(workingBitmap, value));
        }
    }
    return resultList;
}

/**
 * Fetches the requested value...
 */
template<class T>
bool GeneralDataReader::
_getValueForAllConfigs(const std::string& columnName,
                       const std::vector<std::string>& selectionValues,
                       T& resultValue)
{
    for (uint32_t idx = 0; idx < _namedConfigLineage.size(); ++idx) {
        if (_getValueForSingleConfig(columnName,
                                     _namedConfigLineage[idx],
                                     selectionValues,
                                     resultValue)) {
            return true;
        }
    }
    return false;
}

/**
 * Fetches the requested value...
 */
template<class T>
bool GeneralDataReader::
_getValueForSingleConfig(const std::string& columnName,
                        const uint64_t& configId,
                        std::vector<std::string> selectionValues,
                        T& resultValue)
{
    if (_getValueFromDB(columnName, configId,
                        selectionValues, resultValue)) {
        return true;
    }
    while (selectionValues.size() > 0) {
        selectionValues.pop_back();
        //if (selectionValues.size() == 0 && _globalDefault.size() > 0) {
        //    resultValue = boost::lexical_cast<T>(_globalDefault);
        //    return true;
        //}
        //else {
            bool found = _getDefaultValue(columnName, configId,
                                          selectionValues,
                                          resultValue);
            //if (_globalDefault.size() == 0 && selectionValues.size() == 0) {
            //    _globalDefault = boost::lexical_cast<std::string>(resultValue);
            //}
            if (found) {return true;}
            //}
    }
    return false;
}

/**
 * Fetches the requested value...
 */
template<class T>
bool GeneralDataReader::
_getDefaultValue(const std::string& columnName,
                 const uint64_t& configId,
                 const std::vector<std::string>& selectionValues,
                 T& resultValue)
{
    return _getValueFromDB(columnName, configId,
                           selectionValues, resultValue);
}

/**
 * Fetches the requested value...
 */
template<class T>
bool GeneralDataReader::
_getValueFromDB(const std::string& columnName,
                const uint64_t& configId,
                const std::vector<std::string>& selectionValues,
                T& resultValue)
{
    if (! _hasValidConnection) {return false;}
    _dbTablePtr->Clear();
    if (! _setValidityRanges(selectionValues)) {
        return false;
    }
    _dbTablePtr->SetValidityRange("subsysCfgId", configId);
    if (_dbTablePtr->LoadFromDB()) {
        if (_dbTablePtr->NRow() == 1) {
            NOVADB::Row* rowPtr = _dbTablePtr->GetRow(0);
            NOVADB::Column* colPtr = rowPtr->Col(columnName);
            if (colPtr == 0) {return false;}
            return colPtr->Get(resultValue);
        }
    }
    return false;
}

/**
 * Fetches the requested values...
 */
template<class T>
bool GeneralDataReader::
_getListOfValuesForAllConfigs(const std::string& columnName,
                              const std::vector<std::string>& selectionValues,
                              const uint32_t& fineGrainedSelectionLowValue,
                              const uint32_t& fineGrainedSelectionHighValue,
                              std::vector<T>& resultValues)
{
    // require there to be an exact match between the number of
    // selection columns and the number of selection values
    // (including the fine-grained one)
    if (selectionValues.size() != (_selectionColumns.size() - 1)) {
        return false;
    }

    // verify that the requested number of values is reasonable
    if (fineGrainedSelectionLowValue > fineGrainedSelectionHighValue) {
        return false;
    }

    // if the requested number of values is one, simply call the
    // single-value method?

    // initialize the working list of values
    SimpleDataSet<T> dataSet;
    dataSet.numberOfValidValues = 0;
    for (uint32_t rdx = fineGrainedSelectionLowValue;
         rdx <= fineGrainedSelectionHighValue; ++rdx) {
        boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
        dataPtr.reset(new SimpleDataContainer<uint32_t, T>());
        dataPtr->key = rdx;
        dataPtr->hasValue = false;
        dataSet.dataMap[rdx] = dataPtr;
    }

    // loop over named configurations until we've found all of
    // the requested values
    for (uint32_t idx = 0; idx < _namedConfigLineage.size(); ++idx) {
        if (! _getListOfValuesForSingleConfig(columnName,
                                              _namedConfigLineage[idx],
                                              selectionValues,
                                              fineGrainedSelectionLowValue,
                                              fineGrainedSelectionHighValue,
                                              dataSet)) {
            return false;
        }
        if (dataSet.isComplete()) {
            for (uint32_t rdx = fineGrainedSelectionLowValue;
                 rdx <= fineGrainedSelectionHighValue; ++rdx) {
                boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
                dataPtr = dataSet.dataMap[rdx];
                resultValues.push_back(dataPtr->value);
            }
            return true;
        }
    }
    return false;
}

/**
 * Fetches the requested values...
 */
template<class T>
bool GeneralDataReader::
_getListOfValuesForSingleConfig(const std::string& columnName,
                                const uint64_t& configId,
                                std::vector<std::string> selectionValues,
                                const uint32_t& fineGrainedSelectionLowValue,
                                const uint32_t& fineGrainedSelectionHighValue,
                                SimpleDataSet<T>& dataSet)
{
    if (! _getListOfValuesFromDB(columnName, configId,
                                 selectionValues,
                                 fineGrainedSelectionLowValue,
                                 fineGrainedSelectionHighValue,
                                 dataSet)) {
        return false;
    }
    if (dataSet.isComplete()) {return true;}

    selectionValues.push_back("bogusEntry");
    while (selectionValues.size() > 0) {
        selectionValues.pop_back();
        T resultValue;
        bool found = _getDefaultValue(columnName, configId,
                                      selectionValues,
                                      resultValue);
        if (found) {
            for (uint32_t rdx = fineGrainedSelectionLowValue;
                 rdx <= fineGrainedSelectionHighValue; ++rdx) {
                boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
                dataPtr = dataSet.dataMap[rdx];
                if (! dataPtr->hasValue) {
                    dataPtr->value = resultValue;
                    dataPtr->hasValue = true;
                    ++dataSet.numberOfValidValues;
                }
            }
            break;
        }
    }
    return true;
}

/**
 * Fetches the requested values...
 */
template<class T>
bool GeneralDataReader::
_getListOfValuesFromDB(const std::string& columnName,
                       const uint64_t& configId,
                       std::vector<std::string> selectionValues,
                       const uint32_t& fineGrainedSelectionLowValue,
                       const uint32_t& fineGrainedSelectionHighValue,
                       SimpleDataSet<T>& dataSet)
{
    if (! _hasValidConnection) {return false;}
    _dbTablePtr->Clear();
    if (! _setValidityRanges(selectionValues, false)) {
        return false;
    }
    std::string fineGrainedSelectionColumn =
        _selectionColumns[_selectionColumns.size() - 1];
    _dbTablePtr->SetValidityRange(fineGrainedSelectionColumn,
                                  fineGrainedSelectionLowValue,
                                  fineGrainedSelectionHighValue);
    _dbTablePtr->SetValidityRange("subsysCfgId", configId);
    //_dbTablePtr->SetVerbosity(1);
    if (_dbTablePtr->LoadFromDB()) {
        for (int32_t idx = 0; idx < _dbTablePtr->NRow(); ++idx) {
            NOVADB::Row* rowPtr = _dbTablePtr->GetRow(idx);
            T resultValue;
            NOVADB::Column* colPtr = rowPtr->Col(columnName);
            if (! colPtr->Get(resultValue)) {
                return false;
            }

            int32_t selectionValue;
            colPtr = rowPtr->Col(fineGrainedSelectionColumn);
            if (! colPtr->Get(selectionValue)) {
                return false;
            }

            boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
            dataPtr = dataSet.dataMap[selectionValue];
            if (! dataPtr->hasValue) {
                dataPtr->value = resultValue;
                dataPtr->hasValue = true;
                ++dataSet.numberOfValidValues;
            }
        }
        return true;
    }
    return false;
}

/**
 * Fetches the requested values...
 */
template<class T>
bool GeneralDataReader::
_getDataSetForAllConfigs(const std::string& columnName,
                         const std::vector<std::string>& selectionValues,
                         const uint32_t& fineGrainedSelectionLowValue,
                         const uint32_t& fineGrainedSelectionHighValue,
                         SimpleDataSet<T>& resultDataSet)
{
    // require there to be an exact match between the number of
    // selection columns and the number of selection values
    // (including the fine-grained one)
    if (selectionValues.size() != (_selectionColumns.size() - 1)) {
        return false;
    }

    // verify that the requested number of values is reasonable
    if (fineGrainedSelectionLowValue > fineGrainedSelectionHighValue) {
        return false;
    }

    // if the requested number of values is one, simply call the
    // single-value method?

    // initialize the resulting data set
    resultDataSet.dataMap.clear();
    resultDataSet.numberOfValidValues = 0;
    for (uint32_t rdx = fineGrainedSelectionLowValue;
         rdx <= fineGrainedSelectionHighValue; ++rdx) {
        boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
        dataPtr.reset(new SimpleDataContainer<uint32_t, T>());
        dataPtr->key = rdx;
        dataPtr->hasValue = false;
        resultDataSet.dataMap[rdx] = dataPtr;
    }

    // loop over named configurations until we've found all of
    // the requested values
    for (uint32_t idx = 0; idx < _namedConfigLineage.size(); ++idx) {
        if (! _getListOfValuesFromDB(columnName,
                                     _namedConfigLineage[idx],
                                     selectionValues,
                                     fineGrainedSelectionLowValue,
                                     fineGrainedSelectionHighValue,
                                     resultDataSet)) {
            return false;
        }
    }
    return true;
}

/**
 * Fetches the requested values...
 */
template<class T>
bool GeneralDataReader::
_getSetOfValuesForAllConfigs(const std::string& columnName,
                             const std::vector<std::string>& selectionValues,
                             const uint32_t& firstLevelSelectionLowValue,
                             const uint32_t& firstLevelSelectionHighValue,
                             const uint32_t& secondLevelSelectionLowValue,
                             const uint32_t& secondLevelSelectionHighValue,
                             std::vector<boost::shared_ptr<std::vector<T> > >& resultValues)
{
    // require there to be an exact match between the number of
    // selection columns and the number of selection values
    // (including the multiple fine-grained ones)
    if (selectionValues.size() != (_selectionColumns.size() - 2)) {
        return false;
    }

    // verify that the requested number of values is reasonable
    if (firstLevelSelectionLowValue > firstLevelSelectionHighValue) {
        return false;
    }
    if (secondLevelSelectionLowValue > secondLevelSelectionHighValue) {
        return false;
    }

    // if the requested number of values is one, simply call the
    // single-value method?

    // initialize the working list of values
    std::map<uint32_t, boost::shared_ptr<SimpleDataSet<T> > > twoLevelDataSet;
    for (uint32_t idx = firstLevelSelectionLowValue;
         idx <= firstLevelSelectionHighValue; ++idx) {
        boost::shared_ptr<SimpleDataSet<T> > dataSetPtr;
        dataSetPtr.reset(new SimpleDataSet<T>());
        dataSetPtr->numberOfValidValues = 0;
        twoLevelDataSet[idx] = dataSetPtr;
        for (uint32_t jdx = secondLevelSelectionLowValue;
             jdx <= secondLevelSelectionHighValue; ++jdx) {
            boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
            dataPtr.reset(new SimpleDataContainer<uint32_t, T>());
            dataPtr->key = jdx;
            dataPtr->hasValue = false;
            dataSetPtr->dataMap[jdx] = dataPtr;
        }
    }

    // loop over named configurations until we've found all of
    // the requested values
    for (uint32_t idx = 0; idx < _namedConfigLineage.size(); ++idx) {
        if (! _getSetOfValuesForSingleConfig(columnName,
                                             _namedConfigLineage[idx],
                                             selectionValues,
                                             firstLevelSelectionLowValue,
                                             firstLevelSelectionHighValue,
                                             secondLevelSelectionLowValue,
                                             secondLevelSelectionHighValue,
                                             twoLevelDataSet)) {
            return false;
        }

        bool totallyComplete = true;
        for (uint32_t idx = firstLevelSelectionLowValue;
             idx <= firstLevelSelectionHighValue; ++idx) {
            boost::shared_ptr<SimpleDataSet<T> > dataSetPtr;
            dataSetPtr = twoLevelDataSet[idx];
            if (! dataSetPtr->isComplete()) {
                totallyComplete = false;
                break;
            }
        }
        if (totallyComplete) {
            for (uint32_t idx = firstLevelSelectionLowValue;
                 idx <= firstLevelSelectionHighValue; ++idx) {
                boost::shared_ptr<SimpleDataSet<T> > dataSetPtr;
                dataSetPtr = twoLevelDataSet[idx];
                boost::shared_ptr<std::vector<T> > secondLevelValues;
                secondLevelValues.reset(new std::vector<T>());

                for (uint32_t jdx = secondLevelSelectionLowValue;
                     jdx <= secondLevelSelectionHighValue; ++jdx) {
                    boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
                    dataPtr = dataSetPtr->dataMap[jdx];
                    secondLevelValues->push_back(dataPtr->value);
                }

                resultValues.push_back(secondLevelValues);
            }
            return true;
        }
    }
    return false;
}

/**
 * Fetches the requested values...
 */
template<class T>
bool GeneralDataReader::
_getSetOfValuesForSingleConfig(const std::string& columnName,
                               const uint64_t& configId,
                               std::vector<std::string> selectionValues,
                               const uint32_t& firstLevelSelectionLowValue,
                               const uint32_t& firstLevelSelectionHighValue,
                               const uint32_t& secondLevelSelectionLowValue,
                               const uint32_t& secondLevelSelectionHighValue,
                               std::map<uint32_t, boost::shared_ptr<
                               SimpleDataSet<T> > >& twoLevelDataSet)
{
    if (! _getSetOfValuesFromDB(columnName, configId,
                                selectionValues,
                                firstLevelSelectionLowValue,
                                firstLevelSelectionHighValue,
                                secondLevelSelectionLowValue,
                                secondLevelSelectionHighValue,
                                twoLevelDataSet)) {
        return false;
    }
    bool totallyComplete = true;
    for (uint32_t idx = firstLevelSelectionLowValue;
         idx <= firstLevelSelectionHighValue; ++idx) {
        boost::shared_ptr<SimpleDataSet<T> > dataSetPtr;
        dataSetPtr = twoLevelDataSet[idx];
        if (! dataSetPtr->isComplete()) {
            totallyComplete = false;
            break;
        }
    }
    if (totallyComplete) {return true;}

    for (uint32_t idx = firstLevelSelectionLowValue;
         idx <= firstLevelSelectionHighValue; ++idx) {
        boost::shared_ptr<SimpleDataSet<T> > dataSetPtr;
        dataSetPtr = twoLevelDataSet[idx];

        selectionValues.push_back(boost::lexical_cast<std::string>(idx));
        T resultValue;
        bool found = _getDefaultValue(columnName, configId,
                                      selectionValues,
                                      resultValue);
        if (found) {
            for (uint32_t jdx = secondLevelSelectionLowValue;
                 jdx <= secondLevelSelectionHighValue; ++jdx) {
                boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
                dataPtr = dataSetPtr->dataMap[jdx];
                if (! dataPtr->hasValue) {
                    dataPtr->value = resultValue;
                    dataPtr->hasValue = true;
                    ++dataSetPtr->numberOfValidValues;
                }
            }
        }
        selectionValues.pop_back();
    }

    totallyComplete = true;
    for (uint32_t idx = firstLevelSelectionLowValue;
         idx <= firstLevelSelectionHighValue; ++idx) {
        boost::shared_ptr<SimpleDataSet<T> > dataSetPtr;
        dataSetPtr = twoLevelDataSet[idx];
        if (! dataSetPtr->isComplete()) {
            totallyComplete = false;
            break;
        }
    }
    if (totallyComplete) {return true;}

    selectionValues.push_back("bogusEntry");
    while (selectionValues.size() > 0) {
        selectionValues.pop_back();
        T resultValue;
        bool found = _getDefaultValue(columnName, configId,
                                      selectionValues,
                                      resultValue);
        if (found) {
            for (uint32_t idx = firstLevelSelectionLowValue;
                 idx <= firstLevelSelectionHighValue; ++idx) {
                boost::shared_ptr<SimpleDataSet<T> > dataSetPtr;
                dataSetPtr = twoLevelDataSet[idx];

                for (uint32_t jdx = secondLevelSelectionLowValue;
                     jdx <= secondLevelSelectionHighValue; ++jdx) {
                    boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
                    dataPtr = dataSetPtr->dataMap[jdx];
                    if (! dataPtr->hasValue) {
                        dataPtr->value = resultValue;
                        dataPtr->hasValue = true;
                        ++dataSetPtr->numberOfValidValues;
                    }
                }
            }
            break;
        }
    }
    return true;
}

/**
 * Fetches the requested values...
 */
template<class T>
bool GeneralDataReader::
_getSetOfValuesFromDB(const std::string& columnName,
                      const uint64_t& configId,
                      std::vector<std::string> selectionValues,
                      const uint32_t& firstLevelSelectionLowValue,
                      const uint32_t& firstLevelSelectionHighValue,
                      const uint32_t& secondLevelSelectionLowValue,
                      const uint32_t& secondLevelSelectionHighValue,
                      std::map<uint32_t, boost::shared_ptr<
                      SimpleDataSet<T> > >& twoLevelDataSet)
{
    if (! _hasValidConnection) {return false;}
    _dbTablePtr->Clear();
    if (! _setValidityRanges(selectionValues, false)) {
        return false;
    }
    std::string firstLevelSelectionColumn =
        _selectionColumns[_selectionColumns.size() - 2];
    _dbTablePtr->SetValidityRange(firstLevelSelectionColumn,
                                  firstLevelSelectionLowValue,
                                  firstLevelSelectionHighValue);
    std::string secondLevelSelectionColumn =
        _selectionColumns[_selectionColumns.size() - 1];
    _dbTablePtr->SetValidityRange(secondLevelSelectionColumn,
                                  secondLevelSelectionLowValue,
                                  secondLevelSelectionHighValue);
    _dbTablePtr->SetValidityRange("subsysCfgId", configId);
    //_dbTablePtr->SetVerbosity(1);
    if (_dbTablePtr->LoadFromDB()) {
        for (int32_t idx = 0; idx < _dbTablePtr->NRow(); ++idx) {
            NOVADB::Row* rowPtr = _dbTablePtr->GetRow(idx);

            T resultValue;
            NOVADB::Column* colPtr = rowPtr->Col(columnName);
            if (! colPtr->Get(resultValue)) {
                return false;
            }

            int32_t firstSelectionValue;
            colPtr = rowPtr->Col(firstLevelSelectionColumn);
            if (! colPtr->Get(firstSelectionValue)) {
                return false;
            }

            int32_t secondSelectionValue;
            colPtr = rowPtr->Col(secondLevelSelectionColumn);
            if (! colPtr->Get(secondSelectionValue)) {
                return false;
            }

            boost::shared_ptr<SimpleDataSet<T> > dataSetPtr;
            dataSetPtr = twoLevelDataSet[firstSelectionValue];

            boost::shared_ptr<SimpleDataContainer<uint32_t, T> > dataPtr;
            dataPtr = dataSetPtr->dataMap[secondSelectionValue];
            if (! dataPtr->hasValue) {
                dataPtr->value = resultValue;
                dataPtr->hasValue = true;
                ++dataSetPtr->numberOfValidValues;
            }
        }
        return true;
    }
    return false;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils

#endif
