///
/// \file    NearDetDAQChannelMap.h
/// \brief   Channel map specific to the 3x3 near detector
/// \version $Id: NearDetDAQChannelMap.cpp,v 1.1 2014/06/17 19:28:53 messier Exp $
/// \author  Hayes Merritt <hdmerrit@indiana.edu>
///
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include "DAQChannelMap/NearDetDAQChannelMap.h"
#include "DAQChannelMap/NearDetDAQChannelMapConstants.h"
#include "DAQChannelMap/BitFields.h"
#include "DAQChannelMap/boundscheck.h"
#include "NovaDAQConventions/DAQConventions.h"
using namespace daqchannelmap;

NearDetDAQChannelMap::NearDetDAQChannelMap() { }

//......................................................................

NearDetDAQChannelMap::~NearDetDAQChannelMap() { }

//......................................................................

dchan NearDetDAQChannelMap::encodeDChan(lchan logicalchan) const
{
  const int       detector = getDetector(logicalchan);
  const plane_t   plane    = getPlane(logicalchan);
  const cell_t    cell     = getCell(logicalchan);
  const diblock_t diblock  = (plane/ND_NUM_PLANES_DIBLOCK) + 1;
  
  //
  // Initialize the fiber id to an illegal value
  //
  fiber_t fiber = -1;

  //
  // Fiber and cell differ only for horizontal modules (even planes)
  //
  if (plane%2==0) {
    //
    // Fiber calculation for all diblocks excluding the Muon Catcher
    // (diblock<4)
    //
    if (diblock<4) {
      fiber = ND_NUM_CELLS_PLANE - 1 - cell;
    }
    else {
      //
      // Fiber calculation for Muon Catcher. MC is 2 modules tall
      // hence ND_NUM_CELLS_XPLANE_MC replaces ND_NUM_CELLS_PLANE.
      //
      fiber = ND_NUM_CELLS_XPLANE_MC - 1 - cell;
    }
  }
  else {
    //
    // Vertical modules, simple case
    //
    fiber = cell;
  }
  
  dcm_id_t dcm   = computeDCM(plane, fiber, diblock);
  feb_t    feb   = computeFEB(plane, fiber, diblock);
  pixel_t  pixel = computePixel(fiber);

  return DAQChannelMap::encodeDChan(detector, diblock, dcm, feb, pixel);
}

//......................................................................

lchan NearDetDAQChannelMap::encodeLChan(dchan daqchan) const
{
  const int       detector = getDetector(daqchan);
  const diblock_t diblock  = getDiBlock(daqchan);
  const dcm_id_t  dcm      = getDCM(daqchan);
  const feb_t     feb      = getFEB(daqchan);
  const pixel_t   pixel    = getPixel(daqchan);
  
  const plane_t plane = computePlane(diblock, dcm, feb);
  const cell_t  cell  = computeCell(diblock, dcm, feb, pixel);
  
  return DAQChannelMap::encodeLChan(detector, plane, cell);
}

//......................................................................

DetView_TYPE NearDetDAQChannelMap::getView(dchan d) const
{
  if (getDiBlock(d)<4) {
    //
    // Upstream of muon catcher, 4 DCMs, 2 on top, 2 on side
    //
    if (getDCM(d)<3) return X_VIEW;
    else             return Y_VIEW;
  }
  else {
    //
    // In muon catcher, 2 DCMs 1 on top, 1 on side
    //
    if (getDCM(d)==1) return X_VIEW;
    else              return Y_VIEW;
  }
  //
  // All legal cases handled above -- should never reach here.
  //
  abort();
  return INIT;
}

//......................................................................
//
// DCM number calculation using global plane number and fiber number.
//
dcm_id_t NearDetDAQChannelMap::computeDCM(plane_t plane, 
					  fiber_t fiber,
					  diblock_t diblock) const
{
  //
  // Handle diblocks upstream of the muon catcher (diblock<4)
  //
  if (diblock < 4){
    //
    // Even planes are horizontal modules
    //
    if (plane%2==0){      
      dcm_id_t dcm_id = 0;
      if      (fiber< 64) dcm_id = 3;
      else if (fiber>=64) dcm_id = 4;
      return dcm_id;
    }
    else {
      //
      // plane is odd corresponding to vertical modules.
      //
      dcm_id_t dcm_id = 0;
      if      (fiber< 32) dcm_id = 1;
      else if (fiber>=32) dcm_id = 2;
      return dcm_id;
    }
  }
  else {
    //
    // Handle the muon catcher here
    //
    if (plane%2 == 0){
      //
      // Plane is event, horizontal modules
      //
      dcm_id_t dcm_id = 2;
      return dcm_id;
    }
    else {
      //
      // Plane is odd corresponding to vertical modules.
      //
      dcm_id_t dcm_id = 1;
      return dcm_id;
    }
  }
  return 0;
}

//......................................................................

feb_t NearDetDAQChannelMap::computeFEB(plane_t plane,
				       fiber_t fiber,
				       diblock_t diblock) const
{
  feb_t feb = -1;

  if (diblock < 4) {
    //
    // First handle diblocks upstream of the muon catcher (diblock<4)
    //
    if (plane%2==0) {
      //
      // Plane is even corresponding to horizontal modules.
      //
      if (fiber<32) feb = (plane%64)/2 + 32;
      else          feb = (plane%64)/2;
    }
    else {
      //
      // plane is odd corresponding to vertical modules.
      //
      if (fiber < 64) feb = (plane%64)/2;
      else            feb = (plane%64)/2 + 32;
    }
  }
  else if(diblock==4) {
    //
    // Diblock 4 is the Muon Catcher.  FEB calculation is slightly
    // different when compared to other diblocks.
    //
    if (plane%2!=0){
      //
      // Plane is odd corresponding to vertical modules.
      //
      if      (fiber<32)              feb = (plane%64)/2;
      else if (fiber>=32 && fiber<64) feb = (plane%64)/2 + 11;
      else if (fiber>=64)             feb = (plane%64)/2 + 22;
    }
    else {
      //
      // Plane is even corresponding to horizontal modules.
      //
      if (fiber<32) feb = (plane%64)/2;
      else          feb = (plane%64)/2 + 11;
    }
  }
  return feb;
}

//......................................................................

pixel_t NearDetDAQChannelMap::computePixel(fiber_t fiber) const
{
  fiber_t local_fiber = fiber%32;
  pixel_t pixel = FiberToPixel(local_fiber);
  return pixel;
}

//......................................................................

plane_t NearDetDAQChannelMap::computePlane(diblock_t diblock,
					   dcm_id_t dcm,
					   feb_t feb) const
{
  //
  // Handle diblocks upstream of the muon catcher (diblock<4)
  //
  if (diblock < 4) {
    if (dcm < 3) {
      //
      // Vertical modules
      //
      plane_t plane = (feb % 32)*2 + 1 + (diblock-1)*64;
      return plane;
    }
    else {
      //
      // Horizontal modules.
      //
      plane_t plane = (feb % 32)*2 + (diblock-1)*64;
      return plane;
    }
  }
  else {
    //
    // feb number is modded with 11 because there are 11 planes in
    // each view in the Muon Catcher.
    //
    if (dcm==1) {
      plane_t plane = (feb % 11)*2 + 1 + (diblock-1)*64;
      return plane;
    }
    else {
      plane_t plane = (feb % 11)*2 + (diblock-1)*64;
      return plane;
    }
  }
  return 0;
}

//......................................................................

cell_t NearDetDAQChannelMap::computeCell(diblock_t diblock,
					 dcm_id_t  dcm,
					 feb_t     feb,
					 pixel_t   pixel) const
{
  //
  // Handle diblocks upsteam of the muon catcher (diblock<4)
  //
  if (diblock < 4) {
    if (dcm < 3) {
      //
      // Vertical modules.
      //
      cell_t cell = ((dcm - 1)*32 + (feb/32)*32 + PixelToFiber(pixel));
      return cell;
    }
    else {
      //
      // Horizontal modules.
      //
      cell_t cell = 
	(ND_NUM_CELLS_PLANE-1) - 
	((dcm - 3)*64 + (1- feb/32 - dcm/4)*32 + PixelToFiber(pixel));
      return cell;
    }
  }
  else {
    //
    // In Muon Catcher.
    //
    if (dcm==1) {
      cell_t cell = (feb/11)*32 + PixelToFiber(pixel);
      return cell;
    }
    else {
      cell_t cell = 
	(ND_NUM_CELLS_XPLANE_MC-1) - 
	((feb/11)*32 + PixelToFiber(pixel));
      return cell;
    } 
  }
  return 0;
}

//......................................................................

block_t NearDetDAQChannelMap::computeBlock(lchan logicalchan) const
{
  plane_t plane = getPlane(logicalchan);
  return plane/ND_NUM_PLANES_BLOCK;
}

//......................................................................

fiber_t NearDetDAQChannelMap::PixelToFiber(pixel_t pixel) const
{
  return pixel2fiber[pixel];
}

//......................................................................

pixel_t NearDetDAQChannelMap::FiberToPixel(fiber_t fiber) const
{
  return fiber2pixel[fiber];
}

//......................................................................

plane_t NearDetDAQChannelMap::computePlaneInBlock(lchan logicalchan) const
{
  plane_t plane = getPlane(logicalchan);

  // Blocks in the near detector are constructed as 24 plane objects
  plane_t plane_in_block = plane % 24;
  return plane_in_block;
}

//......................................................................

unsigned int NearDetDAQChannelMap::computeModule(lchan logicalchan) const
{
  cell_t cell = getCell(logicalchan);
  return cell/32;
}

//......................................................................

fiber_t NearDetDAQChannelMap::computeFiberInModule(dchan daqchan) const
{
  pixel_t pixel = getPixel(daqchan);
  fiber_t fiber = PixelToFiber(pixel);
  
  return fiber;
}

//......................................................................

unsigned int NearDetDAQChannelMap::getTotalNumberOfDiblocks() const
{
  return ND_NUM_DIBLOCKS;
}

//......................................................................

unsigned int NearDetDAQChannelMap::getNumberOfDCM(DiBlock_TYPE /*dbt*/) const
{
  return ND_NUM_DCMS;
}

//......................................................................

DiBlock_TYPE NearDetDAQChannelMap::getDiBlockType(dchan chan) const
{
  unsigned diblock = getDiBlock(chan);
  if (diblock<4) return AFULLAFULL_DIBLOCK;
  return MC_ND_DIBLOCK;
}

////////////////////////////////////////////////////////////////////////
