#include "DAQDataFormats/RawMicroBlockV0.h"
#include "DAQDataFormats/RawMilliSliceHeader.h"
#include <boost/format.hpp>

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawMicroBlock
#define  CURRENT_CLASS_VERSION        0
#define  CURRENT_FIELDS_SUB_NAMESPACE rawmicroblock
#include "DAQDataFormats/DataFormatLast.inc"

namespace daqdataformats{
namespace VERSION_NAMESPACE{

//------------------------------------------------------------------------------
// Default constructor
RawMicroBlock::RawMicroBlock(const version_t version):
     RawDAQData(version, true, (const sizeofdata_t) &RawMicroBlock::getPredefinedSize)
    ,_Header     (NULL)
    ,_MicroSlice (NULL)
{ }

//------------------------------------------------------------------------------
RawMicroBlock::RawMicroBlock(const RawMicroBlock& copyin):
     RawDAQData(copyin)
    ,_Header     (NULL)
    ,_MicroSlice (NULL)
{ }

//------------------------------------------------------------------------------
// Destructor
RawMicroBlock::~RawMicroBlock() {
    if(_Header)     delete _Header;
    if(_MicroSlice) delete _MicroSlice;
}

//------------------------------------------------------------------------------
bool RawMicroBlock::setBufferSource(const void* pointer){

        PRINT_ON_DEBUG(__PRETTY_FUNCTION__<<" initial size of internal buffer"<<_InternalBuffer.size())

        _Buffer = (void*)pointer;

        _shouldBufferBeInternal = isInternalBuffer();

        /// If Header is already defined, then set its buffer source
        /// Otherwise, need to construct it
        if(_Header) {
            _Header->setBufferSource(pointer);
            const uint32_t* pp = (uint32_t*)pointer + _Header->sizeofdata();
            return _MicroSlice->setBufferSource(pp);
        }
        PRINT_ON_DEBUG(__PRETTY_FUNCTION__<<" step0 size of internal buffer "<<_InternalBuffer.size()<<"  000000000000000000000000000000000000000000000000000\n")

        /// Construct header
        _Header = new daqdataformats::RawMicroBlockHeader(_Buffer, _version); ///< version of the MicroBlockHeader equals the version of the MicroBlock

        /// If buffer is internal, resize it to header size plus one,
        /// so that we can assige the start of the MicroSlice buffer
        if(_shouldBufferBeInternal){
            resizeInternalBuffer(_Header->sizeofdata() + 1);
            /// Check pointers after each read or resize buffer
            checkBufferInternalness();
        }

        /// Pointer to the MicroSlice buffer
        const uint32_t* pp = (uint32_t*)pointer + _Header->sizeofdata();
        
        /// Construct the MicroSlice
        _MicroSlice        = new RawMicroSlice((void*)pp);

        PRINT_ON_DEBUG(__PRETTY_FUNCTION__<<" step1 size of internal buffer "<<_InternalBuffer.size()<<"  1111111111111111111111111111111111111111111111111111\n")

        EXECUTE_ON_DEBUG(
                         if(_function_sizeofdata)std::cout<<"sizeofdata before binding "<<sizeofdata()<<"\n";
                         else                    std::cout<<"sizeofdata is not bound\n";
                        )

        _function_sizeofdata = (sizeofdata_t) &version0::RawMicroBlock::sizeofdata;

        PRINT_ON_DEBUG("sizeofdata after binding "<<sizeofdata())
        return true;
    }

//------------------------------------------------------------------------------
// Printing the MicroBlock structure
void RawMicroBlock::print(std::ostream& os) const {

    const uint32_t size_of_data = EXECUTE_FUNCTION(sizeofdata)();

    os<<boost::format("\n---MicroBlock---------------------------------------------(size=%u, version=%u)\n")
            % size_of_data % _version;

    printBuffer(os);

    for(uint32_t i = 0; i < size_of_data; ++i){
            os<<" ";
            printWord(i, false, os);
            os<<"\n";
    }

    _Header->print(os);
    _MicroSlice->print(os);
    printBinary(_Buffer, _Header->sizeofdata(), os);
}

//------------------------------------------------------------------------------
// Size of the MicroBlock _Buffer in words
uint32_t RawMicroBlock::sizeofdata() const {

    /// We know that the header and MicroSlice is constructed
    if(_MicroSlice->getByteCount() == 0)
        return _Header->sizeofdata();
    
    return _Header->sizeofdata() + _MicroSlice->sizeofdata();
}

//------------------------------------------------------------------------------
// This is a general method to read the microslice from millislice_buffer to _Buffer
void* RawMicroBlock::readData(const void* microblock_buffer)
{
    /// Read the header and find out the size of the MicroBlock Header and MicroSlice Header
    void* buffer_new = readDataGeneral(microblock_buffer, MICROBLOCK_HEADER_STUFF, 0);

    /// Check pointers after each read or resize buffer
    checkBufferInternalness();

    /// Now, read the rest of the microblock
    buffer_new = readDataGeneral(buffer_new, EXECUTE_FUNCTION(sizeofdata)() - MICROBLOCK_HEADER_STUFF, MICROBLOCK_HEADER_STUFF);

    /// Check pointers after each read or resize buffer
    checkBufferInternalness();

    return buffer_new;

}// end of readData(void*)

//------------------------------------------------------------------------------
bool RawMicroBlock::setMicroSlice(const void* microslice_buffer) {

    const daqdataformats::RawMicroSliceHeader header(microslice_buffer);

    const uint32_t micro_size = header.getMicroSliceSize();

    if(_shouldBufferBeInternal){

            resizeInternalBuffer(micro_size);
            /// Check pointers after each read or resize buffer
            checkBufferInternalness();
    }

    readDataGeneral(microslice_buffer, micro_size, _Header->sizeofdata());

    return true;
}


}}// end of namespaces
