#include <DatabaseUtils/DAQAppMgr/ConfiguredProcessList.h>
#include <DatabaseUtils/DAQAppMgr/ApplicationType.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <DatabaseUtils/DAQConfig/DBColumnUtils.h>
#include <NovaDatabase/Table.h>
#include <NovaDatabase/Util.h>
#include <NovaDAQUtilities/EnvVarCache.h>
#include <boost/shared_ptr.hpp>
#include <boost/algorithm/string.hpp>

namespace NOVADB = nova::database;
namespace DAQCFG = dbutils::daqconfig;

namespace dbutils {
namespace daqappmgr {

/**
 * Creates a new instance based on the specified detector name
 * (NDOS, NearDet, FarDet).
 */
ConfiguredProcessList::ConfiguredProcessList(const std::string& detectorName,
                                             int partitionNumber,
                                             const bool& useDBData)
{
    _hasValidData = false;

    if (useDBData) {
        std::string fname;

        fname = "NovaDatabase/tables/DAQAppMgr/BNEVBApplicationMap.xml";
        if (! _fetchProcessData(fname, detectorName,
                                ApplicationType::BNEVB_APPTYPE)) {return;}

        fname = "NovaDatabase/tables/DAQAppMgr/ControlApplicationMap.xml";
        if (! _fetchProcessData(fname, detectorName, "")) {return;}

        fname = "NovaDatabase/tables/DAQAppMgr/DataLoggerApplicationMap.xml";
        if (! _fetchProcessData(fname, detectorName,
                                ApplicationType::DL_APPTYPE)) {return;}

        fname = "NovaDatabase/tables/DAQAppMgr/DCMApplicationMap.xml";
        if (! _fetchProcessData(fname, detectorName,
                                ApplicationType::DCMAPP_APPTYPE)) {return;}
    }
}

/**
 * Fetches the list of applications with the specified type.
 */
bool ConfiguredProcessList::
getProcessListByType(const std::string& appType,
                     std::vector<ApplicationInstance>& resultingAppList) const
{
    if (! _hasValidData) {return false;}

    for (int idx = 0; idx < (int) _fullProcessList.size(); ++idx) {
        if (_fullProcessList[idx].getTypeName() == appType) {
            resultingAppList.push_back(_fullProcessList[idx]);
        }
    }

    return true;
}

/**
 * Fetches the list of applications that do not have one of the
 * specified types.
 */
bool ConfiguredProcessList::
getProcessListByExcludedType(const std::vector<std::string>& appTypeList,
                             std::vector<ApplicationInstance>&
                             resultingAppList) const
{
    if (! _hasValidData) {return false;}

    for (int idx = 0; idx < (int) _fullProcessList.size(); ++idx) {
        bool found = false;
        for (int jdx = 0; jdx < (int) appTypeList.size(); ++jdx) {
            if (_fullProcessList[idx].getTypeName() == appTypeList[jdx]) {
                found = true;
                break;
            }
        }
        if (! found) {
            resultingAppList.push_back(_fullProcessList[idx]);
        }
    }

    return true;
}

/**
 * Fetches the specified application, by type and name.
 */
boost::shared_ptr<ApplicationInstance> ConfiguredProcessList::
getProcess(const std::string& appType, const std::string& appName) const
{
    boost::shared_ptr<ApplicationInstance> appInstancePtr;
    if (! _hasValidData) {return appInstancePtr;}

    for (int idx = 0; idx < (int) _fullProcessList.size(); ++idx) {
        if (_fullProcessList[idx].getTypeName() == appType &&
            _fullProcessList[idx].getAppName() == appName) {
            appInstancePtr.reset(new ApplicationInstance(_fullProcessList[idx]));
            break;
        }
    }

    return appInstancePtr;
}

/**
 * Fetches the process data from the specified database table.
 */
bool ConfiguredProcessList::
_fetchProcessData(const std::string& tableDefinitionFilename,
                  const std::string& detectorName,
                  const std::string& inputAppType)
{
    boost::shared_ptr<NOVADB::Table>
        dbt(new NOVADB::Table(tableDefinitionFilename.c_str()));

    std::string detName = 
      novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    std::string dbUser =
      novadaq::EnvVarCache::getInstance().getEnvVar("CFGMGR_DB_USER");
    if (dbUser.length() > 0) {
      dbt->SetUser(dbUser);
    }
    
    dbt->Clear();
    dbt->SetValidityRange("detectorName", detectorName);
    dbt->SetValidityRange("tstart", NOVADB::Util::GetFarPastTimeAsString(),
                          NOVADB::Util::GetCurrentTimeAsString());
    dbt->SetValidityRange("tstop", NOVADB::Util::GetCurrentTimeAsString(),
                          NOVADB::Util::GetFarFutureTimeAsString());
    //dbt->SetVerbosity(1);
    if (! dbt->LoadFromDB()) {return false;}

    std::vector<std::string> colNames = dbt->GetColNames();
    for (int idx = 0; idx < dbt->NRow(); ++idx) {
        NOVADB::Row* rowPtr = dbt->GetRow(idx);
        NOVADB::Column* colPtr;

        std::string appName;
        std::string appType = inputAppType;
        std::string appHost;
        bool isEnabled;
        boost::shared_ptr<SimpleParameterSet>
            paramSet(new SimpleParameterSet());

        for (int colIdx = 0; colIdx < (int) colNames.size(); ++colIdx) {
            std::string colName = boost::to_lower_copy(colNames[colIdx]);
            if (DAQCFG::DBColumnUtils::isInfrastructureColumn(colName)) {
                continue;
            }

            std::string valueString;
            colPtr = rowPtr->Col(colName);
            if (! colPtr->Get(valueString)) {
                return false;
            }

            if (colName == "appname") {
                appName = valueString;
            }
            else if (colName == "apptype") {
                appType = valueString;
            }
            else if (colName == "host") {
                appHost = valueString;
            }
            else if (colName == "enabled") {
                if (valueString.find("t") != std::string::npos ||
                    valueString.find("T") != std::string::npos ||
                    valueString.find("1") != std::string::npos) {
                    isEnabled = true;
                }
                else {
                    isEnabled = false;
                }
            }

            paramSet->put(colName, valueString);
        }

        boost::shared_ptr<ApplicationType> typePtr;
        boost::shared_ptr<Host> hostPtr;
        ApplicationInstance appInstance(appName, appType, appHost,
                                        typePtr, hostPtr, isEnabled);
        appInstance.setParameterSet(paramSet);
        _addApplicationInstance(appInstance);
    }

    return true;
}

/**
 * Stores the information about the specified application instance.
 */
void ConfiguredProcessList::
_addApplicationInstance(const ApplicationInstance& appInstance)
{
    _hasValidData = true;

    _fullProcessList.push_back(appInstance);

    if (std::find(_uniqueAppTypes.begin(), _uniqueAppTypes.end(),
                  appInstance.getTypeName()) == _uniqueAppTypes.end()) {
        _uniqueAppTypes.push_back(appInstance.getTypeName());
    }

    if (std::find(_uniqueHosts.begin(), _uniqueHosts.end(),
                  appInstance.getHostName()) == _uniqueHosts.end()) {
        _uniqueHosts.push_back(appInstance.getHostName());
    }
}

};  // end of namespace daqappmgr
};  // end of namespace dbutils
