#include <DatabaseUtils/DAQConfig/DCMConnectConfig.h>
#include <DatabaseUtils/DAQConfig/ConfigDataTree.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>

namespace dbutils {
namespace daqconfig {

/**
 * Creates a DCM connect configuration object from the specified
 * list of subsystem configuration IDs.
 *
 * @param subsystemConfigIdList The list of subsystem
 *        configuration IDs.
 * @param dcmName The name of the DCM of interest.
 */
DCMConnectConfig::
DCMConnectConfig(const std::vector<int64_t>& subsystemConfigIdList,
                 const std::string& dcmName)
{
    _primarySubsystemConfigId = -1;
    if (subsystemConfigIdList.size() > 0) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList, dcmName);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a DCM connect configuration object from the specified
 * subsystem configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 * @param dcmName The name of the DCM of interest.
 */
DCMConnectConfig::
DCMConnectConfig(const SubsystemNameSpec& nameSpec,
                 const std::string& dcmName)
{
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestry(nameSpec, subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
       _hasValidData = _fetchAllData(subsystemConfigIdList, dcmName);
    }
    else {
        _hasValidData = false;
    }
}

/**
 * Creates a DCM connect configuration object from the specified
 * global configuration specification.
 *
 * @param nameSpec The named configuration of interest.
 * @param dcmName The name of the DCM of interest.
 */
DCMConnectConfig::
DCMConnectConfig(const GlobalNameSpec& nameSpec,
                 const std::string& dcmName)
{
    _hasValidData = false;
    _primarySubsystemConfigId = -1;
    int64_t subsystemConfigId;
    if (NamedConfigUtils::
        getSubsystemConfigIdFromGlobalName(nameSpec,
                                           SubsystemNameSpec::
                                           DCM_SUBSYSTEM_NAME,
                                           SubsystemNameSpec::
                                           CONNECT_CONFIG_TYPE,
                                           subsystemConfigId)) {
        std::vector<int64_t> subsystemConfigIdList;
        IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM, IDSpec::DAQ);
        if (NamedConfigUtils::
            getConfigIdAncestry(idSpec, subsystemConfigIdList)) {
            _primarySubsystemConfigId = subsystemConfigIdList[0];
            _hasValidData = _fetchAllData(subsystemConfigIdList, dcmName);
        }
    }
}

/**
 * Creates a DCM connect configuration object from the specified
 * global configuration ID.
 *
 * @param globalConfigId The ID of the global configuration of interest.
 * @param dcmName The name of the DCM of interest.
 */
DCMConnectConfig::
DCMConnectConfig(const int64_t& globalConfigId,
                 const std::string& dcmName)
{
    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL, IDSpec::DAQ);
    _primarySubsystemConfigId = -1;
    std::vector<int64_t> subsystemConfigIdList;
    if (NamedConfigUtils::
        getSubsystemConfigIdAncestryFromGlobalId(idSpec,
                                                 SubsystemNameSpec::
                                                 DCM_SUBSYSTEM_NAME,
                                                 SubsystemNameSpec::
                                                 CONNECT_CONFIG_TYPE,
                                                 subsystemConfigIdList)) {
        _primarySubsystemConfigId = subsystemConfigIdList[0];
        _hasValidData = _fetchAllData(subsystemConfigIdList, dcmName);
    }
    else {
        _hasValidData = false;
    }
}

bool DCMConnectConfig::
_fetchAllData(const std::vector<int64_t>& subsystemConfigIdList,
              const std::string& dcmName)
{
    bool status = true;

    // dcm-specific data
    if (status) {
        status = _fetchDCMApplicationParams(subsystemConfigIdList, dcmName);
    }

    return status;
}

bool DCMConnectConfig::
_fetchDCMApplicationParams(const std::vector<int64_t>& subsystemConfigIdList,
                           const std::string& dcmName)
{
    ConfigDataSelector cfgSelector;
    cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                   dcmName, dcmName);

    std::vector<std::string> dataColumnsOfInterest;
    dataColumnsOfInterest.push_back("sendTimeoutInMilliSec");
    dataColumnsOfInterest.push_back("sendNumAttempt");
    dataColumnsOfInterest.push_back("sendSocketBufferSize");

    ConfigDataTree
        dcmParamDataSet("NovaDatabase/tables/DAQConfig/DCMApplicationConnectParameters.xml",
                        subsystemConfigIdList, cfgSelector,
                        dataColumnsOfInterest);
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "sendTimeoutInMilliSec",
                                                   _sendTimeoutInMilliSec)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "sendNumAttempt",
                                                   _sendNumAttempt)) {
        return false;
    }
    if (! dcmParamDataSet.getEffectiveValueFor1Key(dcmName,
                                                   "sendSocketBufferSize",
                                                   _sendSocketBufferSize)) {
        return false;
    }
    return true;
}

};  // end of namespace daqconfig
};  // end of namespace dbutils
