#include "DatabaseUtils/DAQConfig/GlobalNameSpec.h"
#include "DatabaseUtils/DAQConfig/NamedConfigUtils.h"
#include <NovaDAQUtilities/EnvVarCache.h>
#include "NovaDatabase/Table.h"
#include "NovaDatabase/Util.h"

using namespace nova::database;
using namespace dbutils::daqconfig;

/**
 * Command-line utility for updating a global configuration record
 * to show that changes are done.
 */

int main(int argc, char **argv)
{
    // parse command line arguments and print usage hints, if needed.
    std::string timestamp = "";
    bool printUsage = false;
    int c;
    while ((c = getopt (argc, argv, "t:h")) != -1) {
        switch (c)
        {
        case 't':
            timestamp = optarg;
            break;
        case 'h':
            printUsage = true;
            break;
        default:
            printUsage = true;
            break;
        }
    }

    // fetch the command line arguments
    std::string daqOrSlowControls;
    int64_t globalConfigId = 0;
    std::string configName;
    std::string dcmOpMode;
    std::string detectorName;
    if (! printUsage) {
        if ((argc - optind) < 2) {
            printUsage = true;
        }
        else {
            {
                std::string tmpString(argv[optind]);
                if (tmpString == "DAQ" || tmpString == "SlowControls") {
                    daqOrSlowControls = tmpString;
                }
                else {
                    printUsage = true;
                }
                ++optind;
            }
            try {
                std::string firstArg(argv[optind]);
                globalConfigId = boost::lexical_cast<int64_t>(firstArg);
                ++optind;
            }
            catch (...) {
                if ((argc - optind) < 3) {
                    printUsage = true;
                }
                else {
                    {
                        std::string tmpString(argv[optind]);
                        configName = tmpString;
                        ++optind;
                    }
                    {
                        std::string tmpString(argv[optind]);
                        dcmOpMode = tmpString;
                        ++optind;
                    }
                    {
                        std::string tmpString(argv[optind]);
                        detectorName = tmpString;
                        ++optind;
                    }
                }
            }
        }
    }

    // print usage hints, if needed.
    if (printUsage) {
        std::cout << "Usage: " << argv[0] << " [-t '<timestamp>'] "
                  << "<\"DAQ\"|\"SlowControls\"> <globalCfgName> <dcmOpMode> "
                  << "<detectorName>" << std::endl;
        std::cout << " OR    " << argv[0] << " "
                  << "<\"DAQ\"|\"SlowControls\"> <globalCfgId>" << std::endl;
        std::cout << "where \"DAQ\"|\"SlowControls\" selects which set "
                  << "of database tables to use" << std::endl;
        std::cout << "      globalCfgName specifies the named "
                  << "configuration to show" << std::endl;
        std::cout << "      dcmOpMode specifies the DCM operating mode "
                  << "(e.g. DCS)" << std::endl;
        std::cout << "      detectorName specifies the detector name "
                  << "(e.g. NDOS)" << std::endl;
        std::cout << "      globalCfgId specifies the "
                  << "named global configuration to show" << std::endl;
        std::cout << "     -t specifies a particular time to select (UTC)"
                  << std::endl;
        return -1;
    }

    // tell the user which database he/she is using
    std::string dbName =
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBNAME");
    std::cout << "*** You are currently using the " << dbName
              << " database. ***" << std::endl;

    // fetch the details of the user-specified global configuration
    std::vector<NamedConfigUtils::NamedConfigDetailsPtr> detailsList;
    if (globalConfigId != 0) {
        IDSpec idSpec(globalConfigId, IDSpec::GLOBAL,
                      (daqOrSlowControls == "DAQ" ? IDSpec::DAQ :
                       IDSpec::SLOWCONTROLS));
        detailsList = NamedConfigUtils::getConfigDetailAncestry(idSpec);
    }
    else if (timestamp.length() < 14) {
        GlobalNameSpec globalNameSpec(configName, dcmOpMode, detectorName);
        globalNameSpec.isDAQConfiguration(daqOrSlowControls == "DAQ");
        globalNameSpec.skipInvalidStateTest(true);
        detailsList =
            NamedConfigUtils::getGlobalConfigDetailAncestry(globalNameSpec);
    }
    else {
        GlobalNameSpec globalNameSpec(configName, dcmOpMode, detectorName,
                                      timestamp);
        globalNameSpec.isDAQConfiguration(daqOrSlowControls == "DAQ");
        globalNameSpec.skipInvalidStateTest(true);
        detailsList =
            NamedConfigUtils::getGlobalConfigDetailAncestry(globalNameSpec);
    }
    if (detailsList.size() == 0) {
        std::cerr << "*** ERROR: Unable to fetch the database IDs "
                  << "for the specified input parameters! "
                  << "(File = " << __FILE__ << ", Line = "
                  << __LINE__ << ")" << std::endl;
        return 2;
    }

    // fetch parameter values for the global config
    std::vector<int64_t> globalConfigIdsThatNeedFixed;
    NamedConfigUtils::NamedConfigDetailsPtr detailsPtr = detailsList[0];
    std::string idString;
    if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                          NamedConfigUtils::
                                          ID_COLUMN_NAME,
                                          idString)) {
        std::cerr << "*** ERROR: Unable to fetch the global "
                  << "configuration ID. (File = " << __FILE__
                  << ", Line = " << __LINE__ << ")" << std::endl;
        return 2;
    }
    try {
        globalConfigId = boost::lexical_cast<int64_t>(idString);
    }
    catch (...) {
        std::cerr << "*** ERROR: Unable to decode the global "
                  << "configuration ID. (File = " << __FILE__
                  << ", Line = " << __LINE__ << ")" << std::endl;
        return 2;
    }
    if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                          NamedConfigUtils::
                                          NAME_COLUMN_NAME,
                                          configName)) {
        std::cerr << "*** ERROR: Unable to fetch the global "
                  << "configuration name. (File = " << __FILE__
                  << ", Line = " << __LINE__ << ")" << std::endl;
        return 2;
    }
    std::string insertTime;
    if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                          NamedConfigUtils::
                                          INSERTTIME_COLUMN_NAME,
                                          insertTime)) {
        std::cerr << "*** ERROR: Unable to fetch the global "
                  << "configuration insert time. (File = " << __FILE__
                  << ", Line = " << __LINE__ << ")" << std::endl;
        return 2;
    }
    std::string isChangingFlag;
    if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                          NamedConfigUtils::
                                          ISCHANGING_COLUMN_NAME,
                                          isChangingFlag)) {
        std::cerr << "*** ERROR: Unable to fetch the global "
                  << "configuration \"isChanging\" flag. (File = " << __FILE__
                  << ", Line = " << __LINE__ << ")" << std::endl;
        return 2;
    }
    std::string indent = "";
    std::cout << indent << "* " << configName
              << ", id = " << idString
              << ", time = '" << insertTime << "' (UTC)"
              << std::endl;
    if (isChangingFlag == "1") {
        std::cout << indent << "      *** Configuration " << globalConfigId
                  << " is currently marked as having a change in "
                  << "progress. ***" << std::endl;
        globalConfigIdsThatNeedFixed.push_back(globalConfigId);
    }

    // fetch the subsystem configurations associated with the global config
    IDSpec globalIdSpec(globalConfigId, IDSpec::GLOBAL,
                        (daqOrSlowControls == "DAQ" ? IDSpec::DAQ :
                         IDSpec::SLOWCONTROLS));
    std::vector<int64_t> subsystemConfigIdsThatNeedFixed;
    std::vector<NamedConfigUtils::NamedConfigDetailsPtr>
        subsystemDetailsList = NamedConfigUtils::
        getSubsystemConfigsFromGlobalId(globalIdSpec);

    for (int idx = 0; idx < (int) subsystemDetailsList.size(); ++idx) {
        detailsPtr = subsystemDetailsList[idx];
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::ID_COLUMN_NAME,
                                              idString)) {
            std::cerr << "*** ERROR: Unable to fetch the subsystem "
                      << "configuration ID. (File = " << __FILE__
                      << ", Line = " << __LINE__ << ")" << std::endl;
            return 2;
        }
        int64_t subsystemConfigId;
        try {
            subsystemConfigId = boost::lexical_cast<int64_t>(idString);
        }
        catch (...) {
            std::cerr << "*** ERROR: Unable to decode the subsystem "
                      << "configuration ID. (File = " << __FILE__
                      << ", Line = " << __LINE__ << ")" << std::endl;
            return 2;
        }
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              NAME_COLUMN_NAME,
                                              configName)) {
            std::cerr << "*** ERROR: Unable to fetch the subsystem "
                      << "configuration name. (File = " << __FILE__
                      << ", Line = " << __LINE__ << ")" << std::endl;
            return 2;
        }
        std::string subsystemName;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              SUBSYSTEMNAME_COLUMN_NAME,
                                              subsystemName)) {
            std::cerr << "*** ERROR: Unable to fetch the subsystem "
                      << "configuration ID. (File = " << __FILE__
                      << ", Line = " << __LINE__ << ")" << std::endl;
            return 2;
        }
        std::string configType;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              CONFIGTYPE_COLUMN_NAME,
                                              configType)) {
            std::cerr << "*** ERROR: Unable to fetch the subsystem "
                      << "configuration ID. (File = " << __FILE__
                      << ", Line = " << __LINE__ << ")" << std::endl;
            return 2;
        }
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              ISCHANGING_COLUMN_NAME,
                                              isChangingFlag)) {
            std::cerr << "*** ERROR: Unable to fetch the subsystem "
                      << "configuration \"isChanging\" flag. (File = "
                      << __FILE__ << ", Line = " << __LINE__ << ")"
                      << std::endl;
            return 2;
        }
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              INSERTTIME_COLUMN_NAME,
                                              insertTime)) {
            std::cerr << "*** ERROR: Unable to fetch the subsystem "
                      << "configuration creation time. (File = " << __FILE__
                      << ", Line = " << __LINE__ << ")" << std::endl;
            return 2;
        }

        std::cout << indent << "  - " << subsystemName
                  << " " << configType << " " << configName
                  << ", id = " << idString
                  << ", time = '" << insertTime
                  << "' (UTC)" << std::endl;
        if (isChangingFlag == "1") {
            std::cout << indent << "      *** Configuration "
                      << subsystemConfigId
                      << " is currently marked as having a change in "
                      << "progress. ***" << std::endl;
            subsystemConfigIdsThatNeedFixed.push_back(subsystemConfigId);
        }
    }

    // exit if nothing needs fixed
    if (globalConfigIdsThatNeedFixed.size() == 0 && 
        subsystemConfigIdsThatNeedFixed.size() == 0) {
        std::cout << std::endl;
        std::cout << "Exiting: All of the configurations that were found are "
                  << "already marked as done." << std::endl;
        return 0;
    }

    // prompt user whether to continue
    std::cout << std::endl;
    std::cout << "Do you want to mark the configurations listed above "
              << "as done (yes/no): ";
    std::string response;
    std::cin >> response;
    if (response.length() == 0 ||
        (response[0] != 'y' && response[0] != 'Y')) {
        return 1;
    }
    std::cout << std::endl;
    std::cout << "Have you verified that the timestamps are far enough "
              << "in the past " << std::endl
              << "  to ensure that you won't disturb a change "
              << "in progress (yes/no): ";
    std::cin >> response;
    if (response.length() == 0 ||
        (response[0] != 'y' && response[0] != 'Y')) {
        return 1;
    }

    // update the database
    for (int idx=0; idx<(int)globalConfigIdsThatNeedFixed.size(); ++idx) {
        IDSpec idSpec(globalConfigIdsThatNeedFixed[idx], IDSpec::GLOBAL,
                      (daqOrSlowControls == "DAQ" ?
                       IDSpec::DAQ : IDSpec::SLOWCONTROLS));
        if (! NamedConfigUtils::markConfigDone(idSpec, false)) {
            std::cerr << "*** ERROR: Unable to mark global "
                      << "configuration "
                      << globalConfigIdsThatNeedFixed[idx]
                      << " done. (File = " << __FILE__
                      << ", Line = " << __LINE__ << ")" << std::endl;
            return 2;
        }
    }
    for (int idx=0; idx<(int)subsystemConfigIdsThatNeedFixed.size(); ++idx) {
        IDSpec idSpec(subsystemConfigIdsThatNeedFixed[idx], IDSpec::SUBSYSTEM,
                      (daqOrSlowControls == "DAQ" ?
                       IDSpec::DAQ : IDSpec::SLOWCONTROLS));
        if (! NamedConfigUtils::markConfigDone(idSpec, false)) {
            std::cerr << "*** ERROR: Unable to mark subsystem "
                      << "configuration "
                      << subsystemConfigIdsThatNeedFixed[idx]
                      << " done. (File = " << __FILE__
                      << ", Line = " << __LINE__ << ")" << std::endl;
            return 2;
        }
    }

    return 0;
}
