#include <NovaDAQUtilities/LocationUtils.h>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/filesystem.hpp>
#include <iostream>
#include <wordexp.h>

namespace BFS = boost::filesystem;

namespace novadaq {

const std::string LocationUtils::SRT_PROTOCOL_PREFIX("srt://");
const std::string LocationUtils::FILE_PROTOCOL_PREFIX("file://");
const std::string LocationUtils::HTTP_PROTOCOL_PREFIX("http://");
const std::string LocationUtils::FTP_PROTOCOL_PREFIX("ftp://");
const std::string LocationUtils::XML_PROTOCOL_PREFIX("xml://");

/**
 * Translates the specified URL to a generally accepted form.
 *
 * At the moment, the only translation that is supported is to
 * convert a URL of the form srt://relativePath/filename to
 * ${SRT_PRIVATE_RELEASE}/relativePath/filename or
 * ${SRT_PUBLIC_RELEASE}/relativePath/filename depending
 * on whether a private (test) release has been set up *and*
 * the specified file or directory exists in the test release.
 * All other URLs are simply returned with no modification.
 *
 * @param sourceURL The URL string to translate.
 *
 * @return The modified URL string.
 */
std::string LocationUtils::translateURL(const std::string& sourceURL) {

    // make a copy of the URL without any leading spaces (just in case)
    std::string trimmedURL = boost::algorithm::trim_left_copy(sourceURL);

    // if the URL uses our SRT notation, do the needed translation
    if (trimmedURL.substr(0, 6) == "srt://") {
        trimmedURL.erase(0, 5);

        char* rootPath = getenv("SRT_PRIVATE_CONTEXT");
        if (rootPath == 0 || strcmp(rootPath, ".") == 0) {
            rootPath = getenv("SRT_PUBLIC_CONTEXT");
        }
        else {
            std::string fileInTestRelease(rootPath);
            fileInTestRelease.append(trimmedURL);
            if (! BFS::exists(fileInTestRelease)) {
                rootPath = getenv("SRT_PUBLIC_CONTEXT");
            }
        }

        if (rootPath != 0) {
            trimmedURL.insert(0, rootPath);
            return trimmedURL;
        }
    }

    // if the URL uses our SRT notation, do the needed translation
    if (trimmedURL.substr(0, 6) == "xml://") {
        trimmedURL.erase(0, 5);

        char* rootPath = getenv("NOVADAQXML_DIR");
        if (rootPath) {
	  trimmedURL.insert(0, rootPath);
	  return trimmedURL;
        }
    }

    // if we failed to do the translation, or none was needed, simply
    // return the input URL
    return sourceURL;
}

/**
 * Replaces environmental variable names in the input string with
 * their values from the current shell environment.
 *
 * *NOTE* that for this replacement to work, environmental
 * variable names must be surrounded by "${" and "}".
 *
 * For example:
 * - ${SRT_PRIVATE_CONTEXT}/${HOST}
 *
 * @param sourceString The input string.
 *
 * @return The modified string.
 */
std::string LocationUtils::replaceEnvVars(const std::string& sourceString) {

    // make a copy of the string
    std::string workString = sourceString;

    // loop through all of the env vars in the string
    std::string::size_type startPos = workString.find("${", 0);
    while (startPos != std::string::npos) {
        std::string::size_type endPos = workString.find("}", startPos+1);
        if (endPos != std::string::npos &&
            endPos > startPos) {
            std::string::size_type length = endPos - startPos - 2;

            std::string envVarName = workString.substr(startPos+2, length);
            std::string toBeReplaced = workString.substr(startPos, length+3);

            char* envVarPtr = getenv(envVarName.c_str());
            if (envVarPtr != 0) {
                std::string replacementString(envVarPtr);
                boost::replace_all(workString, toBeReplaced,
                                   replacementString);
            }
        }

        startPos = workString.find("${", startPos+1);
    }

    return workString;
}

/**
 * Runs the word expansion function on the input string
 * (<code>wordexp</code>).  This function provides tilde
 * expansion and environmental variable replacement, among
 * other things.
 *
 * @param sourceString The input string.
 *
 * @return The modified string.
 */
std::string LocationUtils::wordExpansion(const std::string& sourceString) {

    std::string resultString = sourceString;

    wordexp_t result;
    int status = wordexp(sourceString.c_str(), &result, WRDE_NOCMD);
    if (status == 0) {
        resultString = "";
        for (uint32_t idx = 0; idx < result.we_wordc; ++idx) {
            std::string tmpString(result.we_wordv[idx]);
            resultString.append(tmpString);
        }
    }

    wordfree(&result);
    return resultString;
}

/**
 * Finds the first occurrence of the specified file using a standard
 * set of possible locations.  This method returns true if an instance of
 * the file was found and returns false if not.  The location of the
 * instance of the file that was found is returned in the
 * locationOfFoundInstance argument.
 *
 * NOTE that if the file is found on disk, a "file://" prefix is currently
 * *NOT* included in the returned locationOfFoundInstance string.  It simply
 * contains the full path to the requested file.
 *
 * The current search order is the following:
 * <ol>
 * <li> `pwd`/filePath (if the filePath does *not* have a leading slash)
 * <li> filePath (if the filePath does have a leading slash)
 * <li> srt://filePath (this translates to ${SRT_PRIVATE_CONTEXT}/filePath
 *      followed by ${SRT_PUBLIC_CONTEXT}/filePath)
 * <li> http://filePath (not implemented yet)
 * <li> ftp://filePath (not implemented yet)
 * </ol>
 *
 * If suggested prefixes are supplied, and the searches listed above do
 * not initially succeed in locating the file, then the search is repeated
 * using modifiedFilePath = suggestedPrefix + filePath for each 
 * suggested prefix until the file is found or the list is exhausted.
 * 
 * @param filePath The name of the file to be located.  This string may
 *        contain optional absolute or relative path information.
 * @param locationOfFoundInstance The URL of the first instance of the
 *        file that was found.
 * @param suggestedPrefixes A list of prefixes that should be prepended
 *        to the filePath during searching.
 *
 * @return true if the file was found, false if not.
 */
bool LocationUtils::
findFile(const std::string& filePath,
         std::string& locationOfFoundInstance,
         const std::vector<std::string>& suggestedPrefixes) {

    // first, try the search without using any suggested prefix
    if (LocationUtils::_findFile(filePath, locationOfFoundInstance)) {
        return true;
    }

    // next, try the search with each suggested prefix
    for (uint32_t idx = 0; idx < suggestedPrefixes.size(); ++idx) {
        std::string modifiedFilePath = suggestedPrefixes[idx] + filePath;
        if (LocationUtils::_findFile(modifiedFilePath,
                                     locationOfFoundInstance)) {
            return true;
        }
    }

    // return false to indicate that we couldn't find the file
    return false;
}

/**
 * Finds the first occurrence of the specified file using a standard
 * set of possible locations.  This method returns true if an instance of
 * the file was found and returns false if not.  The location of the
 * instance of the file that was found is returned in the
 * locationOfFoundInstance argument.
 *
 * @see findFile(const std::string&, std::string&, const std::vector<std::string>&)
 */
bool LocationUtils::findFile(const std::string& filePath,
                             std::string& locationOfFoundInstance) {
    std::vector<std::string> emptyList;
    return LocationUtils::findFile(filePath, locationOfFoundInstance,
                                   emptyList);
}

/**
 * Does the work of searching for a file in a pre-defined set of locations.
 */
bool LocationUtils::_findFile(const std::string& filePath,
                              std::string& locationOfFoundInstance) {

    // initialization
    std::string workString;
    std::string expandedPath = wordExpansion(filePath);
    std::string trimmedFilePath = boost::algorithm::trim_copy(expandedPath);

    // look for the file without any special protocol prefix
    if (trimmedFilePath.substr(0, 1) == "/") {
        if (BFS::exists(trimmedFilePath)) {
            locationOfFoundInstance = trimmedFilePath;
            return true;
        }
    }
    else {
        BFS::path currentPath = BFS::current_path();
#if ( BOOST_FILESYSTEM_VERSION == 2 )
        workString = currentPath.file_string();
#else
        workString = currentPath.string();
#endif
        workString.append("/" + trimmedFilePath);
        if (BFS::exists(workString)) {
            locationOfFoundInstance = workString;
            return true;
        }
    }

    // look for the file in SRT directories
    workString = SRT_PROTOCOL_PREFIX + trimmedFilePath;
    workString = translateURL(workString);
    if (BFS::exists(workString)) {
        locationOfFoundInstance = workString;
        return true;
    }

    // look for the file in xml subdir of Offline release of Online packages
    workString = XML_PROTOCOL_PREFIX + trimmedFilePath;
    workString = translateURL(workString);
    if (BFS::exists(workString)) {
        locationOfFoundInstance = workString;
        return true;
    }

    return false;
}

};  // end of namespace novadaq
